@@ -39,13 +39,13 @@
 #include "comgr-compiler.h"
 #include "comgr-cache-bundler-command.h"
 #include "comgr-cache.h"
+#include "comgr-clang-command.h"
 #include "comgr-device-libs.h"
 #include "comgr-diagnostic-handler.h"
 #include "comgr-env.h"
 #include "comgr-spirv-command.h"
 #include "lld/Common/CommonLinkerContext.h"
 #include "lld/Common/Driver.h"
-#include "clang/Basic/Version.h"
 #include "clang/CodeGen/CodeGenAction.h"
 #include "clang/Driver/Compilation.h"
 #include "clang/Driver/DriverDiagnostic.h"
@@ -59,6 +59,7 @@
 #include "clang/FrontendTool/Utils.h"
 #include "llvm/ADT/StringExtras.h"
 #include "llvm/Bitcode/BitcodeWriter.h"
+#include "llvm/IR/Constants.h"
 #include "llvm/IR/LLVMContext.h"
 #include "llvm/IR/Module.h"
 #include "llvm/IR/Verifier.h"
@@ -84,6 +85,7 @@
 #include "llvm/Support/MemoryBuffer.h"
 #include "llvm/Support/Path.h"
 #include "llvm/Support/Signals.h"
+#include "llvm/Support/VirtualFileSystem.h"
 #include "llvm/Support/WithColor.h"
 #include "llvm/TargetParser/Host.h"
 
@@ -552,6 +554,11 @@ SmallString<128> getFilePath(DataObject *Object, StringRef Dir) {
   return Path;
 }
 
+// TODO: Move inputFromFile and outputToFile within AMDGPUCompiler
+//
+// Currently, we only invoke these two methods in the context of AMDGPUCompiler.
+// Moreover, member functions that deal with file I/O should not worry whether
+// the underlying filesystem being used is virtual or real.
 amd_comgr_status_t inputFromFile(DataObject *Object, StringRef Path) {
   ProfilePoint Point("FileIO");
   auto BufOrError = MemoryBuffer::getFile(Path);
@@ -644,7 +651,8 @@ void logArgv(raw_ostream &OS, StringRef ProgramName,
 }
 
 amd_comgr_status_t executeCommand(const Command &Job, raw_ostream &LogS,
-                                  DiagnosticOptions &DiagOpts) {
+                                  DiagnosticOptions &DiagOpts,
+                                  llvm::vfs::FileSystem &FS) {
   TextDiagnosticPrinter DiagClient(LogS, &DiagOpts);
   IntrusiveRefCntPtr<DiagnosticIDs> DiagID(new DiagnosticIDs);
   DiagnosticsEngine Diags(DiagID, &DiagOpts, &DiagClient, false);
@@ -670,9 +678,11 @@ amd_comgr_status_t executeCommand(const Command &Job, raw_ostream &LogS,
 
     std::unique_ptr<CompilerInstance> Clang(new CompilerInstance());
     Clang->setVerboseOutputStream(LogS);
+    Clang->setFileManager(new FileManager(Clang->getFileSystemOpts(), &FS));
     if (!Argv.back()) {
       Argv.pop_back();
     }
+
     if (!CompilerInvocation::CreateFromArgs(Clang->getInvocation(), Argv,
                                             Diags)) {
       return AMD_COMGR_STATUS_ERROR;
@@ -680,7 +690,7 @@ amd_comgr_status_t executeCommand(const Command &Job, raw_ostream &LogS,
     // Internally this call refers to the invocation created above, so at
     // this point the DiagnosticsEngine should accurately reflect all user
     // requested configuration from Argv.
-    Clang->createDiagnostics(&DiagClient, /* ShouldOwnClient */ false);
+    Clang->createDiagnostics(FS, &DiagClient, /* ShouldOwnClient */ false);
     if (!Clang->hasDiagnostics()) {
       return AMD_COMGR_STATUS_ERROR;
     }
@@ -753,11 +763,11 @@ AMDGPUCompiler::executeInProcessDriver(ArrayRef<const char *> Args) {
   IntrusiveRefCntPtr<DiagnosticIDs> DiagID(new DiagnosticIDs);
   DiagnosticsEngine Diags(DiagID, &*DiagOpts, DiagClient);
 
-  ProcessWarningOptions(Diags, *DiagOpts, /*ReportDiags=*/false);
+  ProcessWarningOptions(Diags, *DiagOpts, *OverlayFS, /*ReportDiags=*/false);
 
   Driver TheDriver((Twine(env::getLLVMPath()) + "/bin/clang").str(),
-                   llvm::sys::getDefaultTargetTriple(), Diags);
-  TheDriver.setTitle("AMDGPU Code Object Manager");
+                   llvm::sys::getDefaultTargetTriple(), Diags,
+                   "AMDGPU Code Object Manager", OverlayFS);
   TheDriver.setCheckInputsExist(false);
 
   // Log arguments used to build compilation
@@ -779,7 +789,7 @@ AMDGPUCompiler::executeInProcessDriver(ArrayRef<const char *> Args) {
 
   auto Cache = CommandCache::get(LogS);
   for (auto &Job : C->getJobs()) {
-    CachedCommand C(Job, *DiagOpts, executeCommand);
+    ClangCommand C(Job, *DiagOpts, *OverlayFS, executeCommand);
     if (Cache) {
       if (auto Status = Cache->execute(C, LogS)) {
         return Status;
@@ -875,7 +885,8 @@ amd_comgr_status_t AMDGPUCompiler::removeTmpDirs() {
 #endif
 }
 
-amd_comgr_status_t AMDGPUCompiler::processFile(const char *InputFilePath,
+amd_comgr_status_t AMDGPUCompiler::processFile(DataObject *Input,
+                                               const char *InputFilePath,
                                                const char *OutputFilePath) {
   SmallVector<const char *, 128> Argv = Args;
 
@@ -896,6 +907,12 @@ amd_comgr_status_t AMDGPUCompiler::processFile(const char *InputFilePath,
     Argv.push_back("-save-temps=obj");
   }
 
+  // Add SPIR-V flags
+  for (auto Flag : Input->SpirvFlags) {
+    Argv.push_back("-Xclang");
+    Argv.push_back(Flag);
+  }
+
   Argv.push_back(InputFilePath);
 
   Argv.push_back("-o");
@@ -907,6 +924,12 @@ amd_comgr_status_t AMDGPUCompiler::processFile(const char *InputFilePath,
 amd_comgr_status_t
 AMDGPUCompiler::processFiles(amd_comgr_data_kind_t OutputKind,
                              const char *OutputSuffix) {
+  return processFiles(OutputKind, OutputSuffix, InSet);
+}
+
+amd_comgr_status_t
+AMDGPUCompiler::processFiles(amd_comgr_data_kind_t OutputKind,
+                             const char *OutputSuffix, DataSet *InSet) {
   for (auto *Input : InSet->DataObjects) {
     if (Input->DataKind != AMD_COMGR_DATA_KIND_INCLUDE) {
       continue;
@@ -944,7 +967,7 @@ AMDGPUCompiler::processFiles(amd_comgr_data_kind_t OutputKind,
     auto OutputFilePath = getFilePath(Output, OutputDir);
 
     if (auto Status =
-            processFile(InputFilePath.c_str(), OutputFilePath.c_str())) {
+            processFile(Input, InputFilePath.c_str(), OutputFilePath.c_str())) {
       return Status;
     }
 
@@ -1040,7 +1063,6 @@ amd_comgr_status_t AMDGPUCompiler::addCompilationFlags() {
     break;
   case AMD_COMGR_LANGUAGE_HIP:
     Args.push_back("hip");
-    Args.push_back("-nogpuinc");
     Args.push_back("--offload-device-only");
     Args.push_back("-isystem");
     Args.push_back(ROCMIncludePath.c_str());
@@ -1087,8 +1109,18 @@ amd_comgr_status_t AMDGPUCompiler::addDeviceLibraries() {
     for (auto DeviceLib : getDeviceLibraries()) {
       llvm::SmallString<128> DeviceLibPath = DeviceLibsDir;
       path::append(DeviceLibPath, std::get<0>(DeviceLib));
-      if (auto Status = outputToFile(std::get<1>(DeviceLib), DeviceLibPath)) {
-        return Status;
+      // TODO: We should abstract the logic of deciding whether to use the VFS
+      // or the real file system within inputFromFile and outputToFile.
+      if (UseVFS) {
+        if (!InMemoryFS->addFile(
+                DeviceLibPath, /* ModificationTime */ 0,
+                llvm::MemoryBuffer::getMemBuffer(std::get<1>(DeviceLib)))) {
+          return AMD_COMGR_STATUS_ERROR;
+        }
+      } else {
+        if (auto Status = outputToFile(std::get<1>(DeviceLib), DeviceLibPath)) {
+          return Status;
+        }
       }
     }
   }
@@ -1868,19 +1900,108 @@ amd_comgr_status_t AMDGPUCompiler::linkToExecutable() {
   return amd_comgr_data_set_add(OutSetT, OutputT);
 }
 
+// TODO: Generalize this list to include all -cc1 flags and arguments that are
+//   still valid in a bitcode compilation context
+static inline const std::unordered_set<std::string_view> ValidSpirvFlags{
+    "-fapprox-func",
+    "-fcolor-diagnostics",
+    "-fconvergent-functions",
+    "-fexceptions",
+    "-ffast-math",
+    "-ffinite-math-only",
+    "-ffp-contract=fast",
+    "-ffp-contract=fast-honor-pragmas",
+    "-fgpu-rdc",
+    "-finline-functions",
+    "-fno-signed-zeros",
+    "-fno-rounding-math",
+    "-fno-experimental-relative-c++-abi-vtables",
+    "-fno-autolink",
+    "-freciprocal-math",
+    "-funsafe-math-optimizations",
+    "-fvisibility=hidden",
+    "-O0",
+    "-O1",
+    "-O2",
+    "-O3",
+    "--save-temps"};
+
+amd_comgr_status_t AMDGPUCompiler::extractSpirvFlags(DataSet *BcSet) {
+
+  for (auto *Bc : BcSet->DataObjects) {
+    // Create SPIRV IR Module from Bitcode Buffer
+    SMDiagnostic SMDiag;
+    LLVMContext Context;
+    Context.setDiagnosticHandler(
+        std::make_unique<AMDGPUCompilerDiagnosticHandler>(this->LogS), true);
+
+    auto Mod = getLazyIRModule(
+        MemoryBuffer::getMemBuffer(StringRef(Bc->Data, Bc->Size), "", false),
+        SMDiag, Context, true);
+
+    if (!Mod) {
+      SMDiag.print("SPIR-V Bitcode", LogS, /* ShowColors */ false);
+      return AMD_COMGR_STATUS_ERROR;
+    }
+
+    if (verifyModule(*Mod, &LogS))
+      return AMD_COMGR_STATUS_ERROR;
+
+    // Fetch @llvm.cmdline
+    GlobalVariable *CmdLine = Mod->getNamedGlobal("llvm.cmdline");
+
+    // Return if no @llvm.cmdline
+    if (!CmdLine)
+      return AMD_COMGR_STATUS_SUCCESS;
+
+    if (ConstantDataSequential *CDS =
+            dyn_cast<ConstantDataSequential>(CmdLine->getInitializer())) {
+
+      // Add each valid null-terminated '\0' string to Flags
+      std::string Tmp;
+      StringRef CmdLineRaw = CDS->getRawDataValues();
+      std::stringstream ss(CmdLineRaw.str());
+      while (getline(ss, Tmp, '\0')) {
+        if (Tmp == "--hipstdpar" || Tmp == "-amdgpu-enable-hipstdpar") {
+          Bc->SpirvFlags.push_back("-mllvm");
+          Bc->SpirvFlags.push_back("-amdgpu-enable-hipstdpar");
+        } else if (ValidSpirvFlags.count(Tmp)) {
+          Bc->SpirvFlags.push_back(Saver.save(Tmp.c_str()).data());
+        }
+      }
+    }
+
+    // COV5 required for SPIRV
+    Bc->SpirvFlags.push_back("-mcode-object-version=5");
+
+    if (env::shouldEmitVerboseLogs()) {
+      LogS << "        SPIR-V Flags: " << Bc->Name << "\n";
+      for (auto Flag : Bc->SpirvFlags)
+        LogS << "          " << Flag << "\n";
+    }
+  }
+
+  return AMD_COMGR_STATUS_SUCCESS;
+}
+
 amd_comgr_status_t AMDGPUCompiler::translateSpirvToBitcode() {
+  return translateSpirvToBitcodeImpl(InSet, DataSet::convert(OutSetT));
+}
+
+amd_comgr_status_t
+AMDGPUCompiler::translateSpirvToBitcodeImpl(DataSet *SpirvInSet,
+                                            DataSet *BcOutSet) {
 #ifdef COMGR_DISABLE_SPIRV
-  LogS << "Calling AMDGPUCompiler::translateSpirvToBitcode() not supported "
-    << "Comgr is built with -DCOMGR_DISABLE_SPIRV. Re-build LLVM and Comgr "
-    << "with LLVM-SPIRV-Translator support to continue.\n";
+  LogS << "Calling AMDGPUCompiler::translateSpirvToBitcodeImpl() not "
+       << "supported. Comgr is built with -DCOMGR_DISABLE_SPIRV. Re-build LLVM "
+       << "and Comgr with LLVM-SPIRV-Translator support to continue.\n";
   return AMD_COMGR_STATUS_ERROR;
 #else
   if (auto Status = createTmpDirs()) {
     return Status;
   }
 
-  auto Cache = CommandCache::get(LogS);
-  for (auto *Input : InSet->DataObjects) {
+  for (auto *Input : SpirvInSet->DataObjects) {
 
     if (env::shouldSaveTemps()) {
       if (auto Status = outputToFile(Input, getFilePath(Input, InputDir))) {
@@ -1895,6 +2016,7 @@ amd_comgr_status_t AMDGPUCompiler::translateSpirvToBitcode() {
     SmallString<0> OutBuf;
     SPIRVCommand SPIRV(Input, OutBuf);
 
+    auto Cache = CommandCache::get(LogS);
     amd_comgr_status_t Status;
     if (!Cache) {
       Status = SPIRV.execute(LogS);
@@ -1918,7 +2040,8 @@ amd_comgr_status_t AMDGPUCompiler::translateSpirvToBitcode() {
     Output->setName(std::string(Input->Name) + std::string(".bc"));
     Output->setData(OutBuf);
 
-    if (auto Status = amd_comgr_data_set_add(OutSetT, OutputT)) {
+    if (auto Status =
+            amd_comgr_data_set_add(DataSet::convert(BcOutSet), OutputT)) {
       return Status;
     }
 
@@ -1937,11 +2060,74 @@ amd_comgr_status_t AMDGPUCompiler::translateSpirvToBitcode() {
 #endif
 }
 
+amd_comgr_status_t AMDGPUCompiler::compileSpirvToRelocatable() {
+  if (auto Status = createTmpDirs()) {
+    return Status;
+  }
+
+  for (auto *Input : InSet->DataObjects) {
+    if (Input->DataKind != AMD_COMGR_DATA_KIND_SPIRV)
+      return AMD_COMGR_STATUS_ERROR_INVALID_ARGUMENT;
+  }
+
+  // Translate .spv to .bc
+  amd_comgr_data_set_t TranslatedSpirvT;
+  if (auto Status = amd_comgr_create_data_set(&TranslatedSpirvT))
+    return Status;
+  DataSet *TranslatedSpirv = DataSet::convert(TranslatedSpirvT);
+
+  if (auto Status = translateSpirvToBitcodeImpl(InSet, TranslatedSpirv))
+    return Status;
+
+  // Extract any SPIR-V flags from @llvm.cmdline
+  if (auto Status = extractSpirvFlags(TranslatedSpirv))
+    return Status;
+
+  // Compile bitcode to relocatable
+  if (ActionInfo->IsaName) {
+    if (auto Status = addTargetIdentifierFlags(ActionInfo->IsaName)) {
+      return Status;
+    }
+  }
+
+  if (ActionInfo->ShouldLinkDeviceLibs) {
+    if (auto Status = addDeviceLibraries()) {
+      return Status;
+    }
+  }
+
+  Args.push_back("-c");
+
+  Args.push_back("-mllvm");
+  Args.push_back("-amdgpu-internalize-symbols");
+
+  return processFiles(AMD_COMGR_DATA_KIND_RELOCATABLE, ".o", TranslatedSpirv);
+}
+
 AMDGPUCompiler::AMDGPUCompiler(DataAction *ActionInfo, DataSet *InSet,
                                DataSet *OutSet, raw_ostream &LogS)
     : ActionInfo(ActionInfo), InSet(InSet), OutSetT(DataSet::convert(OutSet)),
       LogS(LogS) {
   initializeCommandLineArgs(Args);
+
+  // Initialize OverlayFS with the real file system which helps redirect
+  // non-VFS reads and writes.
+  OverlayFS = new vfs::OverlayFileSystem(vfs::getRealFileSystem());
+
+  std::optional<bool> VFSStatus = env::shouldUseVFS();
+  if ((VFSStatus.has_value() && *VFSStatus) ||
+      (!VFSStatus.has_value() && ActionInfo->ShouldUseVFS)) {
+    if (env::shouldEmitVerboseLogs()) {
+      LogS << "    File System: VFS\n";
+    }
+    UseVFS = true;
+    InMemoryFS = new vfs::InMemoryFileSystem;
+    OverlayFS->pushOverlay(InMemoryFS);
+  } else {
+    if (env::shouldEmitVerboseLogs()) {
+      LogS << "    File System: Real\n";
+    }
+  }
 }
 
 AMDGPUCompiler::~AMDGPUCompiler() {
