@@ -253,11 +253,8 @@ AMDGPUTargetInfo::AMDGPUTargetInfo(const llvm::Triple &Triple,
 
   MaxAtomicPromoteWidth = MaxAtomicInlineWidth = 64;
   CUMode = !(GPUFeatures & llvm::AMDGPU::FEATURE_WGP);
-
-  for (auto F : {"image-insts", "gws", "vmem-to-lds-load-insts"}) {
-    if (GPUKind != llvm::AMDGPU::GK_NONE)
-      ReadOnlyFeatures.insert(F);
-  }
+  for (auto F : {"image-insts", "gws", "vmem-to-lds-load-insts"})
+    ReadOnlyFeatures.insert(F);
   HalfArgsAndReturns = true;
 }
 
